/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import org.sablecc.sablecc.analysis.*;
import org.sablecc.sablecc.node.*;
import java.util.*;
import java.io.*;

/*
 * ResolveTransformIds
 * 
 * This class computes semantic verifications for alternative transformations
 * in Production section. It makes sure that all transformations specified
 * for an alternative match with their corresponding prod_transform but also
 * the alternative transformation is itself correct.
 */

public class ResolveTransformIds extends DepthFirstAdapter
{

  private ResolveAstIds transformIds;
  private ResolveAltIds altIds;
  private ResolveProdTransformIds prodTransformIds;

  private int nbTransformAlt = 0;
  private int nbTransformProd = 0;
  private String currentAstAlt;
  private String currentAstProd;
  private String currentProd;
  private String currentAlt;
  private String currentAstProdName;

  //This Map contains the type of any term of alternative transformation(altTransform)
  public final Map altTransformElemTypes = new TypedHashMap(
        NodeCast.instance,
        StringCast.instance);

  public final Map mapSimpleTermProdTransformation = new TypedHashMap(
        StringCast.instance,
        ListCast.instance);

  public final Map simpleTermOrsimpleListTermTypes = new TypedHashMap(
        NodeCast.instance,
        StringCast.instance);

  public ResolveTransformIds(ResolveAstIds ast_ids, ResolveAltIds alt_ids, ResolveProdTransformIds p_ids)
  {
    transformIds = ast_ids;
    altIds = alt_ids;
    prodTransformIds = p_ids;
  }

  public ResolveProdTransformIds getProdTransformIds()
  {
    return prodTransformIds;
  }

  public void inAProd(final AProd production)
  {
    nbTransformProd = 0;
    currentProd = (String)altIds.ids.names.get(production);

    if(production.getArrow() != null)
    {
      nbTransformProd = production.getProdTransform().size();
    }
  }

  private LinkedList listCurrentAltGlobal;
  private LinkedList listCurrentAlt;
  private LinkedList listOfListCurrentAlt;

  public void inAParsedAlt(AParsedAlt nodeAlt)
  {
    nbTransformAlt = 0;

    currentAlt = (String)altIds.ids.names.get(nodeAlt);
    listCurrentAltGlobal = (LinkedList)((LinkedList)altIds.alts_elemsGlobal.get(currentAlt)).clone();
    listCurrentAlt = (LinkedList)((LinkedList)altIds.alts_elems.get(currentAlt)).clone();
    listOfListCurrentAlt = (LinkedList)((LinkedList)altIds.alts_elems_list.get(currentAlt)).clone();
  }

  public void inAAltTransform(AAltTransform node)
  {
    if(node.getTerms().size() == 0)
    {
      altTransformElemTypes.put(node, "nothing");
    }
    else
    {
      nbTransformAlt = node.getTerms().size();
    }
  }

  public void outAAltTransform(AAltTransform node)
  {
    if(nbTransformAlt != nbTransformProd)
    {
      error10( node.getRBrace() );
    }
  }

  public void outAParsedAlt(AParsedAlt node)
  {
    lastSimpleTerm = null;

    listCurrentAltGlobal = null;
    listCurrentAlt = null;
    listOfListCurrentAlt = null;
    mapSimpleTermProdTransformation.clear();
  }

  public void outANewTerm(ANewTerm node)
  {
    LinkedList list = (LinkedList)prodTransformIds.prod_transforms.get(currentProd);
    AProdName prodNameNode = (AProdName)node.getProdName();

    currentAstProd = "P" + ResolveIds.name(prodNameNode.getId().getText());
    currentAstProdName = prodNameNode.getId().getText();

    if(prodNameNode.getProdNameTail() != null)
    {
      TId prodNameTailNode = prodNameNode.getProdNameTail();
      currentAstAlt = "A" + ResolveIds.name(prodNameTailNode.getText()) + ResolveIds.name(prodNameNode.getId().getText());
    }
    else
    {
      currentAstAlt = "A" + ResolveIds.name(prodNameNode.getId().getText());
    }

    String currentastProdName;
    if(currentAstProdName.startsWith("P."))
    {
      currentastProdName = currentAstProdName.substring(2);
    }
    else
    {
      currentastProdName = currentAstProdName;
    }

    if(transformIds.ast_prods.get(currentAstProd) == null)
    {
      error15(((AProdName)node.getProdName()).getId(), ((AProdName)node.getProdName()).getId().getText());
    }

    if(transformIds.ast_alts.get(currentAstAlt) == null)
    {
      error14(((AProdName)node.getProdName()).getId(), currentAstAlt);
    }

    int sizeNewTerm = 0;
    if(node.getParams().size() > 0)
    {
      sizeNewTerm = node.getParams().size();
    }

    int sizeAstAlt = 0;
    if( ((AAstAlt)transformIds.ast_alts.get(currentAstAlt) ).getElems() != null)
    {
      sizeAstAlt = ( (LinkedList)((AAstAlt)transformIds.ast_alts.get(currentAstAlt) ).getElems()).size();
    }

    if(sizeNewTerm != sizeAstAlt)
    {
      error17( ((AProdName)node.getProdName()).getId(), currentAstAlt);
    }

    AProdName aProdName = (AProdName)node.getProdName();

    String type = "P" + ResolveIds.name(aProdName.getId().getText());

    altTransformElemTypes.put(node, type);

    AAstAlt astAlt = (AAstAlt)transformIds.ast_alts.get(currentAstAlt);

    if(node.getParams().size() > 0 && astAlt.getElems().size() > 0)
    {
      Object elemsTable[] = astAlt.getElems().toArray();
      Object paramsTable[] = node.getParams().toArray();

      String termType, elemType;

      for(int j=0; j<elemsTable.length; j++)
      {
        termType = (String)altTransformElemTypes.get(paramsTable[j]);
        elemType = (String)transformIds.ast_elemTypes.get(elemsTable[j]);

        if( (((AElem)elemsTable[j]).getUnOp() != null) &&
            ( (((AElem)elemsTable[j]).getUnOp() instanceof AStarUnOp) || (((AElem)elemsTable[j]).getUnOp() instanceof APlusUnOp) ) )
        {
          elemType = "L" + elemType;
        }
        if((!elemType.startsWith("L") && !termType.equals("null") && !termType.equals(elemType)) ||
            (elemType.startsWith("L") && !termType.equals("Lnull") && !termType.equals(elemType)) )
        {
          error21(node.getLPar(), elemType, termType);
        }
      }
    }
  }

  public void outANewListTerm(ANewListTerm node)
  {
    LinkedList list = (LinkedList)prodTransformIds.prod_transforms.get(currentProd);
    AProdName prodNameNode = (AProdName)node.getProdName();

    currentAstProdName = prodNameNode.getId().getText();
    currentAstProd = "P" + ResolveIds.name(currentAstProdName);

    if(prodNameNode.getProdNameTail() != null)
    {
      TId prodNameTailNode = prodNameNode.getProdNameTail();
      currentAstAlt = "A" + ResolveIds.name(prodNameTailNode.getText()) + ResolveIds.name(prodNameNode.getId().getText());
    }
    else
    {
      currentAstAlt = "A" + ResolveIds.name(prodNameNode.getId().getText());
    }

    String currentastProdName;
    if(currentAstProdName.startsWith("P."))
    {
      currentastProdName = currentAstProdName.substring(2);
    }
    else
    {
      currentastProdName = currentAstProdName;
    }

    if(transformIds.ast_prods.get(currentAstProd) == null)
    {
      error15(((AProdName)node.getProdName()).getId(), ((AProdName)node.getProdName()).getId().getText());
    }

    if(transformIds.ast_alts.get(currentAstAlt) == null)
    {
      error14(((AProdName)node.getProdName()).getId(), currentAstAlt);
    }

    int sizeNewTerm = 0;
    if(node.getParams().size() > 0)
    {
      sizeNewTerm = node.getParams().size();
    }

    int sizeAstAlt = 0;
    if( ((AAstAlt)transformIds.ast_alts.get(currentAstAlt) ).getElems() != null)
    {
      sizeAstAlt = ( (LinkedList)((AAstAlt)transformIds.ast_alts.get(currentAstAlt) ).getElems()).size();
    }

    if(sizeNewTerm != sizeAstAlt)
    {
      error17( ((AProdName)node.getProdName()).getId(), currentAstAlt);
    }

    /***********************************************************************************************************************************/
    AProdName aProdName = (AProdName)node.getProdName();

    String type = "P" + ResolveIds.name(aProdName.getId().getText());

    altTransformElemTypes.put(node, type);

    AAstAlt astAlt = (AAstAlt)transformIds.ast_alts.get(currentAstAlt);

    if(node.getParams().size() > 0 && astAlt.getElems().size() > 0)
    {
      Object elemsTable[] = astAlt.getElems().toArray();
      Object paramsTable[] = node.getParams().toArray();

      String termType, elemType;

      for(int j=0; j<elemsTable.length; j++)
      {
        termType = (String)altTransformElemTypes.get(paramsTable[j]);
        elemType = (String)transformIds.ast_elemTypes.get(elemsTable[j]);

        if( (((AElem)elemsTable[j]).getUnOp() != null) &&
            ( (((AElem)elemsTable[j]).getUnOp() instanceof AStarUnOp) || (((AElem)elemsTable[j]).getUnOp() instanceof APlusUnOp) ) )
        {
          elemType = "L" + elemType;
        }
        if((!elemType.startsWith("L") && !termType.equals("null") && !termType.equals(elemType)) ||
            (elemType.startsWith("L") && !termType.equals("Lnull") && !termType.equals(elemType)) )
        {
          error21(node.getLPar(), elemType, termType);
        }
      }
    }
    /***********************************************************************************************************************************/
  }

  public void outAListTerm(AListTerm node)
  {
    if( (node.getListTerms() != null) && (node.getListTerms().size() != 0) )
    {
      Object temp[] = node.getListTerms().toArray();
      String firstTermType = (String)altTransformElemTypes.get(temp[0]);

      for(int i=1; i<temp.length; i++)
      {
        String termType = (String)altTransformElemTypes.get(temp[i]);

        if( ( !termType.startsWith("L") && !firstTermType.startsWith("L") && !termType.equals(firstTermType) ) ||
            ( termType.startsWith("L")  && !firstTermType.startsWith("L") && !(termType.substring(1)).equals(firstTermType) ) ||
            ( !termType.startsWith("L") && firstTermType.startsWith("L")  && !(firstTermType.substring(1)).equals(termType) ) ||
            ( termType.startsWith("L")  && firstTermType.startsWith("L")  && !termType.equals(firstTermType) ) )
        {
          error20(node.getLBkt());
        }
      }

      if(!firstTermType.startsWith("L"))
      {
        altTransformElemTypes.put(node, "L" + firstTermType);
      }
      else
      {
        altTransformElemTypes.put(node, firstTermType);
      }
    }
    else
    {
      altTransformElemTypes.put(node, "Lnull");
    }
  }

  private LinkedList listL;
  private LinkedList listP;
  private String lastSimpleTerm;

  public void inASimpleTerm(ASimpleTerm node)
  {
    String name = node.getId().getText();
    String elemType = (String)transformIds.astIds.altsElemNameTypes.get( currentAlt+"."+node.getId().getText() );
    boolean okTermtail = false;
    String tmpName = name;

    if( (listCurrentAltGlobal == null) || !listCurrentAltGlobal.contains(name) )
    {
      /************************************/error11(node.getId(), name);
    }
    else
    {
      ListIterator iter = null;
      iter = listCurrentAltGlobal.listIterator();

      while(iter.hasNext())
      {
        if( name.equals((String)iter.next()) )
        {
          if( node.getSimpleTermTail() == null )
          {
            iter.remove();
          }
          break;
        }
      }
    }

    if( (listCurrentAlt == null) || !listCurrentAlt.contains(name) )
    {
      error11(node.getId(), name);
    }
    else
    {
      ListIterator iter = null;
      iter = listCurrentAlt.listIterator();

      while(iter.hasNext())
      {
        if( name.equals((String)iter.next()) )
        {
          if( node.getSimpleTermTail() == null )
          {
            iter.remove();
          }
          break;
        }
      }
    }

    int position = 0;

    if(node.getSpecifier() != null)
    {
      if(node.getSpecifier() instanceof ATokenSpecifier)
      {
        if(elemType != null)
        {
          error19(node.getId());
        }

        name = "T" + ResolveIds.name(node.getId().getText());
        //A token can't have term tail
        if(node.getSimpleTermTail() != null)
        {
          error12(node.getId());
        }

        altTransformElemTypes.put(node, "T" + ResolveIds.name(node.getId().getText()) );
        position = 0;
      }
      else
      {
        //the id is an element name
        if(elemType != null)
        {
          error19(node.getId());
        }

        name = "P" + ResolveIds.name(node.getId().getText());

        if( mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName) == null )
        {
          if(prodTransformIds.prod_transforms.get(name) != null)
          {
            listL = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(name)).clone();
            mapSimpleTermProdTransformation.put(currentAlt+"."+tmpName, listL);
          }
        }

        listCurrentAlt.remove(lastSimpleTerm);

        lastSimpleTerm = name;

        boolean blistL = false;

        if( (listL!= null) && (listL.size()==1) && listL.contains(node.getId().getText()) )
        {
          blistL = true;
        }

        //the production is transformed but it appears without a term tail.
        if( (node.getSimpleTermTail() == null) && (listL != null) && !blistL)
        {
          error9(node.getId(), node.getId().getText());
        }

        if( node.getSimpleTermTail() == null )
        {
          position = 0;
        }

        if( node.getSimpleTermTail() != null )
        {
          String strTermTail = node.getSimpleTermTail().getText();

          if( ((listL == null) || !listL.contains( strTermTail ) ) )
          {
            error13(node.getId(), node.getId().getText(), strTermTail);
          }
          else
          {
            ListIterator iter = null;
            iter = listL.listIterator();

            position = listL.indexOf(strTermTail);
            while(iter.hasNext())
            {
              if( strTermTail.equals((String)iter.next()) )
              {
                iter.remove();
                break;
              }
            }
          }
        }
        //The Type of the element without his eventual termtail (term.termtail :: (type of term))
        simpleTermOrsimpleListTermTypes.put(node, "P" + ResolveIds.name(node.getId().getText()));

        if(node.getSimpleTermTail() == null)
        {
          altTransformElemTypes.put(node, "P" + ResolveIds.name(node.getId().getText()) );
        }
        else
        {
          String termTail = node.getSimpleTermTail().getText();
          String typeOfTerm = (String)prodTransformIds.prodTransformElemTypesString.get(name+"."+termTail);
          if(typeOfTerm.startsWith("L"))
          {
            typeOfTerm = typeOfTerm.substring(1);
          }
          altTransformElemTypes.put(node, typeOfTerm) ;
        }
      }
    }
    //The specifier is null
    else
    {
      String type;

      //The element is a token
      if( ( (elemType != null) && elemType.startsWith("T") ) ||
          ( (elemType == null) && transformIds.astIds.tokens.get("T" + ResolveIds.name(node.getId().getText())) != null ) )
      {
        if(node.getSimpleTermTail() != null)
        {
          error12(node.getId());
        }

        if(elemType != null)
        {
          altTransformElemTypes.put(node, elemType);
        }
        else
        {
          altTransformElemTypes.put(node, "T" + ResolveIds.name(node.getId().getText()) );
        }
        position = 0;
      }
      //The element is a production
      else
      {
        listL = null;

        name = "P" + ResolveIds.name(node.getId().getText());

        if(!name.equals(lastSimpleTerm) )
        {
          if(elemType != null)
          {
            if( mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName) == null)
            {
              if(prodTransformIds.prod_transforms.get(elemType) != null)
              {
                listL = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(elemType)).clone();
                mapSimpleTermProdTransformation.put(currentAlt+"."+tmpName, listL);
              }
            }
            else
            {
              listL = (LinkedList)mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName);
            }
          }
          if( mapSimpleTermProdTransformation.get(currentAlt+".P"+tmpName) == null )
          {
            listP = (LinkedList)prodTransformIds.prod_transforms.get(name);
            if(prodTransformIds.prod_transforms.get(name) != null)
            {
              listP = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(name)).clone();
              mapSimpleTermProdTransformation.put(currentAlt+".P"+tmpName, listP);
            }
          }
          else
          {
            listP = (LinkedList)mapSimpleTermProdTransformation.get(currentAlt+".P"+tmpName);
          }

          listCurrentAlt.remove(lastSimpleTerm);
        }

        boolean blistL = false;

        if( ( (elemType != null) && (listL!= null) && (listL.size()==1) && ResolveIds.name((String)listL.getFirst()).equals(elemType.substring(1)) ) ||
            ( (elemType == null) && (listP!= null) && (listP.size()==1) && listP.contains(node.getId().getText()) ) )
        {
          blistL = true;
        }

        //the production is transformed but it appears without a term tail.
        if( (node.getSimpleTermTail() == null) &&
            ( (listL != null) || (listP != null) ) && !blistL )
        {
          if(elemType != null)
          {
            error9(node.getId(), elemType);
          }
          else
          {
            error9(node.getId(), node.getId().getText());
          }
        }

        //simpletermtail doesn't appear in the production transformation
        if( node.getSimpleTermTail() != null )
        {
          okTermtail = true;
          String strTermTail = node.getSimpleTermTail().getText();

          if( ( (listL == null) || !listL.contains(strTermTail) ) &&
              ( (listP == null) || !listP.contains(strTermTail) ) )
          {
            error13(node.getId(), node.getId().getText(), strTermTail);
          }
          else
          {
            ListIterator iter = null;

            if(listL != null)
            {
              iter = listL.listIterator();

              position = listL.indexOf(strTermTail);

              while(iter.hasNext())
              {
                if( strTermTail.equals((String)iter.next()) )
                {
                  iter.remove();
                  break;
                }
              }
            }

            if(listP != null)
            {
              iter = listP.listIterator();

              position = listP.indexOf(strTermTail);

              while(iter.hasNext())
              {
                if( strTermTail.equals((String)iter.next()) )
                {
                  iter.remove();
                  break;
                }
              }
            }

          }
        }

        if(node.getSimpleTermTail() != null)
        {
          String termtail = node.getSimpleTermTail().getText();
          LinkedList listProdContains = null;
          if(elemType != null)
          {
            listProdContains = (LinkedList)prodTransformIds.mapProdTransformContainsList.get(elemType);
          }
          else
          {
            listProdContains = (LinkedList)prodTransformIds.mapProdTransformContainsList.get("P" + ResolveIds.name(node.getId().getText()));
          }
        }
        //The Type of the element without his eventual termtail (term.termtail :: (type of term))
        if(elemType != null)
        {
          simpleTermOrsimpleListTermTypes.put(node, elemType);
        }
        else
        {
          simpleTermOrsimpleListTermTypes.put(node, "P" + ResolveIds.name(node.getId().getText()));
        }

        if(node.getSimpleTermTail() == null)
        {
          if(elemType != null)
          {
            altTransformElemTypes.put(node, elemType);
          }
          else
          {
            altTransformElemTypes.put(node, "P" + ResolveIds.name(node.getId().getText()) );
          }
          position = 0;
        }
        else
        {
          String prodType;
          if(elemType != null)
          {
            prodType = elemType;
          }
          else
          {
            prodType = "P" + ResolveIds.name(node.getId().getText());
          }

          String termTail = node.getSimpleTermTail().getText();
          String typeOfTerm = (String)prodTransformIds.prodTransformElemTypesString.get(prodType+"."+termTail);
          if(typeOfTerm.startsWith("L"))
          {
            typeOfTerm = typeOfTerm.substring(1);
          }
          altTransformElemTypes.put(node, typeOfTerm);
        }
      }
    }
  }

  public void caseANullTerm(ANullTerm node)
  {
    altTransformElemTypes.put(node, "null");
  }

  public void inASimpleListTerm(ASimpleListTerm node)
  {
    String name = node.getId().getText();
    String  elemType = (String)transformIds.astIds.altsElemNameTypes.get( currentAlt+"."+node.getId().getText() );
    String strTermTail;
    String tmpName = name;

    if( (listCurrentAltGlobal == null) || !listCurrentAltGlobal.contains(name) )
    {
      error11(node.getId(), name);
    }
    else
    {
      ListIterator iter = null;
      iter = listCurrentAltGlobal.listIterator();

      while(iter.hasNext())
      {
        if( name.equals((String)iter.next()) )
        {
          if( node.getSimpleTermTail() == null )
          {
            iter.remove();
          }
          break;
        }
      }
    }

    if( ((listCurrentAlt == null) || !listCurrentAlt.contains(name)) && ((listOfListCurrentAlt == null) || !listOfListCurrentAlt.contains(name)) )
    {
      /***********************************************/error11(node.getId(), name);
    }
    else
    {
      ListIterator iter = null;
      if( (listCurrentAlt != null) && listCurrentAlt.contains(name) )
      {
        iter = listCurrentAlt.listIterator();
      }
      else if( (listOfListCurrentAlt != null) && listOfListCurrentAlt.contains(name) )
      {
        iter = listOfListCurrentAlt.listIterator();
      }

      while(iter.hasNext())
      {
        if( name.equals((String)iter.next()) )
        {
          if( node.getSimpleTermTail() == null )
          {
            iter.remove();
          }
          break;
        }
      }
    }

    int position = 0;

    if(node.getSpecifier() != null)
    {
      //The simpleTerm is a token
      if(node.getSpecifier() instanceof ATokenSpecifier)
      {
        //If the token has a name, then it cannot be referred by this name
        //preceded by the specifier "T."
        if(elemType != null)
        {
          error19(node.getId());
        }

        name = "T" + ResolveIds.name(node.getId().getText());

        //this is a simpleTerm which seems to be a token,
        //so it can't have term tail. (like T.id.termtail)
        if(node.getSimpleTermTail() != null)
        {
          error12(node.getId());
        }
        altTransformElemTypes.put(node, "T" + ResolveIds.name(node.getId().getText()) );
        position = 0;
      }
      else
      {
        //the id is an element name
        if(elemType != null)
        {
          error19(node.getId());
        }

        name = "P" + ResolveIds.name(node.getId().getText());

        if( mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName) == null )
        {
          if(prodTransformIds.prod_transforms.get(name) != null)
          {
            listL = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(name)).clone();
            mapSimpleTermProdTransformation.put(currentAlt+"."+tmpName, listL);
          }
        }
        listCurrentAlt.remove(lastSimpleTerm);

        boolean blistL = false;

        if( (listL!= null) && (listL.size()==1) && listL.contains(node.getId().getText()) )
        {
          blistL = true;
        }

        //the production is transformed but it appears without a term tail.
        if( (node.getSimpleTermTail() == null) && (  listL != null ) && !blistL )
        {
          error9(node.getId(), node.getId().getText());
        }

        if( node.getSimpleTermTail() == null )
        {
          position = 0;
        }

        //the production is not transformed but it appears with a term tail.
        if( (node.getSimpleTermTail() != null) )
        {
          strTermTail = node.getSimpleTermTail().getText();

          if (((listL == null) || !listL.contains( strTermTail ) ) )
          {
            error13(node.getId(), node.getId().getText(), strTermTail);
          }
          else
          {
            ListIterator iter = null;
            iter = listL.listIterator();

            position = listL.indexOf(strTermTail);
            while(iter.hasNext())
            {
              if( strTermTail.equals((String)iter.next()) )
              {
                iter.remove();
                break;
              }
            }
          }
        }

        //The Type of the element without his eventual termtail (term.termtail :: (type of term))
        simpleTermOrsimpleListTermTypes.put(node, "P" + ResolveIds.name(node.getId().getText()));

        if(node.getSimpleTermTail() == null)
        {
          altTransformElemTypes.put(node, "P" + ResolveIds.name(node.getId().getText()) );
        }
        else
        {
          String termTail = (String) node.getSimpleTermTail().getText();
          String typeOfTerm = (String)prodTransformIds.prodTransformElemTypesString.get(name+"."+termTail);
          if(typeOfTerm.startsWith("L"))
          {
            typeOfTerm = typeOfTerm.substring(1);
          }
          altTransformElemTypes.put(node, typeOfTerm);
        }
      }
    }

    //the element has no specifier
    else
    {
      String type;
      if( ( (elemType != null) && elemType.startsWith("T") ) ||
          ( (elemType == null) && transformIds.astIds.tokens.get("T" + ResolveIds.name(node.getId().getText())) != null ) )
      {
        //A token can't have term tail
        if(node.getSimpleTermTail() != null)
        {
          error12(node.getId());
        }

        if(elemType != null)
        {
          altTransformElemTypes.put(node, elemType);
        }
        else
        {
          altTransformElemTypes.put(node, "T" + ResolveIds.name(node.getId().getText()) );
        }
        position = 0;
      }
      //it seems to be a production without a specifier
      else
      {
        name = "P" + ResolveIds.name(node.getId().getText());

        if(!name.equals(lastSimpleTerm) )
        {
          if(elemType != null)
          {
            if( mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName) == null)
            {
              if(prodTransformIds.prod_transforms.get(elemType) != null)
              {
                listL = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(elemType)).clone();
                mapSimpleTermProdTransformation.put(currentAlt+"."+tmpName, listL);
              }
            }
            else
            {
              listL = (LinkedList)mapSimpleTermProdTransformation.get(currentAlt+"."+tmpName);
            }
          }
          if( mapSimpleTermProdTransformation.get(currentAlt+".P"+tmpName) == null )
          {
            listP = (LinkedList)prodTransformIds.prod_transforms.get(name);
            if(prodTransformIds.prod_transforms.get(name) != null)
            {
              listP = (LinkedList)((LinkedList)prodTransformIds.prod_transforms.get(name)).clone();
              mapSimpleTermProdTransformation.put(currentAlt+".P"+tmpName, listP);
            }
          }
          else
          {
            listP = (LinkedList)mapSimpleTermProdTransformation.get(currentAlt+".P"+tmpName);
          }

          listCurrentAlt.remove(lastSimpleTerm);
        }

        boolean blistL = false;

        if( ( (elemType != null) && (listL!= null) && (listL.size()==1) && ResolveIds.name((String)listL.getFirst()).equals(elemType.substring(1)) ) ||
            ( (elemType == null) && (listP!= null) && (listP.size()==1) && listP.contains(node.getId().getText()) ) )
        {
          blistL = true;
        }

        //the production is transformed but it appears without a term tail.
        if( (node.getSimpleTermTail() == null) &&
            ( (listL != null) || (listP != null) ) && !blistL )
        {
          if(elemType != null)
          {
            error9(node.getId(), elemType);
          }
          else
          {
            error9(node.getId(), node.getId().getText());
          }
        }

        //the production is not transformed but it appears with a term tail.
        if(node.getSimpleTermTail() != null)
        {
          strTermTail = node.getSimpleTermTail().getText();

          if( ( (listL == null) || !listL.contains(strTermTail) ) &&
              ( (listP == null) || !listP.contains(strTermTail) ) )
          {
            error13(node.getId(), node.getId().getText(), node.getSimpleTermTail().getText() );
          }
          else
          {
            ListIterator iter = null;

            if(listL != null)
            {
              iter = listL.listIterator();

              position = listL.indexOf(strTermTail);

              while(iter.hasNext())
              {
                if( strTermTail.equals((String)iter.next()) )
                {
                  iter.remove();
                  break;
                }
              }
            }

            if(listP != null)
            {
              iter = listP.listIterator();

              position = listP.indexOf(strTermTail);

              while(iter.hasNext())
              {
                if( strTermTail.equals((String)iter.next()) )
                {
                  iter.remove();
                  break;
                }
              }
            }
          }
        }

        if(node.getSimpleTermTail() != null)
        {
          String termtail = node.getSimpleTermTail().getText();
          LinkedList listProdContains = null;
          LinkedList prodContains = null;
          if(elemType != null)
          {
            listProdContains = (LinkedList)prodTransformIds.mapProdTransformContainsList.get(elemType);
            prodContains = (LinkedList)prodTransformIds.prod_transforms.get(elemType);
          }
          else
          {
            listProdContains = (LinkedList)prodTransformIds.mapProdTransformContainsList.get("P" + ResolveIds.name(node.getId().getText()));
            prodContains = (LinkedList)prodTransformIds.prod_transforms.get("P" + ResolveIds.name(node.getId().getText()));
          }

          LinkedList lst = (LinkedList)altIds.alts_elems_list.get(currentAlt);
        }

        //The Type of the element without his eventual termtail (term.termtail :: (type of term))

        if(elemType != null)
        {
          simpleTermOrsimpleListTermTypes.put(node, elemType);
        }
        else
        {
          simpleTermOrsimpleListTermTypes.put(node, "P" + ResolveIds.name(node.getId().getText()));
        }

        if(node.getSimpleTermTail() == null)
        {
          if(elemType != null)
          {
            altTransformElemTypes.put(node, elemType);
          }
          else
          {
            altTransformElemTypes.put(node, "P" + ResolveIds.name(node.getId().getText()) );
          }
          position = 0;
        }
        else
        {
          String prodType;
          if(elemType != null)
          {
            prodType = elemType;
          }
          else
          {
            prodType = "P" + ResolveIds.name(node.getId().getText());
          }

          String termTail = node.getSimpleTermTail().getText();
          String typeOfTerm = (String)prodTransformIds.prodTransformElemTypesString.get(prodType+"."+termTail);
          if(typeOfTerm.startsWith("L"))
          {
            typeOfTerm = typeOfTerm.substring(1);
          }
          altTransformElemTypes.put(node, typeOfTerm);
        }
      }
    }
  }

  private Object temp1[];
  private boolean boolProdTransform;
  private boolean firstProduction = false;

  public void outAProductions(AProductions node)
  {
    LinkedList list = node.getProds();
    AProd prod = (AProd)list.getFirst();
    firstProduction = true;

    if( prodTransformIds.listProdTransformList.contains("P"+ ResolveIds.name( prod.getId().getText()) ) )
    {
      errorFirst(prod.getId());
    }

    Object temp[] = node.getProds().toArray();
    for(int i = 0; i < temp.length; i++)
    {

      ((PProd) temp[i]).apply(new DepthFirstAdapter()
                              {
                                public void inAProd(AProd production)
                                {
                                  LinkedList prodTransform = production.getProdTransform();
                                  String prodTransformElemType = "";
                                  temp1 = null;
                                  boolProdTransform = false;

                                  if(production.getArrow() != null && prodTransform.size() >= 1)
                                  {
                                    temp1 = prodTransform.toArray();
                                    prodTransformElemType = (String)prodTransformIds.prodTransformElemTypes.get(temp1[0]);
                                  }
                                  else if(production.getArrow() == null)
                                  {
                                    boolProdTransform = true;
                                    String []tab = new String[1];
                                    tab[0] = "P" + ResolveIds.name(production.getId().getText());
                                    temp1 = (Object[])tab;
                                    prodTransformElemType = (String)temp1[0];
                                  }
                                  else
                                  {
                                    boolProdTransform = true;
                                    String []tab = new String[1];
                                    tab[0] = "nothing";
                                    temp1 = (Object[])tab;
                                    prodTransformIds.prodTransformElemTypes.put(prodTransform, "nothing");
                                    prodTransformElemType = (String)temp1[0];
                                  }

                                  if(firstProduction)
                                  {
                                    if( !prodTransformElemType.equals(transformIds.getFirstAstProduction()) )
                                    {
                                      error22();
                                    }
                                    firstProduction = false;
                                  }
                                }

                                public void inAAltTransform(AAltTransform node)
                                {
                                  //LinkedList altTransform = node.getAltTransform();
                                  Object temp3[] = node.getTerms().toArray();

                                  if(temp3.length > 0)
                                  {
                                    for(int k = 0; k < temp3.length; k++)
                                    {
                                      String prodTransformElemType, altTransformTermType;

                                      if(!boolProdTransform)
                                      {
                                        prodTransformElemType = (String)prodTransformIds.prodTransformElemTypes.get(temp1[k]);
                                      }
                                      else
                                      {
                                        prodTransformElemType = (String)temp1[k];
                                      }
                                      altTransformTermType = (String)altTransformElemTypes.get(temp3[k]);

                                      if( !(( prodTransformElemType.startsWith("L") && altTransformTermType.equals("Lnull") ||
                                              altTransformTermType.equals(prodTransformElemType)) ||
                                            ( !prodTransformElemType.startsWith("L") && altTransformTermType.equals("null") ||
                                              altTransformTermType.equals(prodTransformElemType) )) )
                                      {
                                        error21(node.getLBrace(), prodTransformElemType, altTransformTermType);
                                      }
                                    }
                                  }
                                  else
                                  {
                                    String prodTransformElemType = (String)temp1[0];
                                    String altTransformTermType = (String)altTransformElemTypes.get(node);
                                    if(!prodTransformElemType.equals(altTransformTermType))
                                    {
                                      error21(node.getLBrace(), prodTransformElemType, altTransformTermType);
                                    }
                                  }
                                }
                              }
                             );
    }
  }

  private static void error9(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "production " + name + " was transformed. It can't appear here again");
  }

  private  void error10(Token token)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "this alternative transformation is incorrect because the number of "+
      " production transformation and alternatives ones doesn't match");
  }

  private static void error11(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      name + " must be one of the elements on the left side of the arrow "+
      " or is already refered to in this alternative" );
  }

  private static void error12(Token token)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "tokens are never transformed. This syntax is incorrect");
  }

  private static void error13(Token token, String prod_name, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "production " + prod_name + " was never transformed to "+name+
      " or #"+name+"# cannot be refered by #" + prod_name + "# twice in the same alternative");
  }

  private static void error14(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "alternative "+ name +" doesn't exist in section AST");
  }

  private static void error15(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "Production "+ name + " doesn't exist in section AST");
  }

  private static void error16(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "this transformation is incorrect. The current production "+
      "can not be transformed to production "+ name);
  }

  private static void error17(Token token, String name)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "The number of parameters request for this alternative in "+
      "section AST doesn't match");
  }

  private static void error19(Token token)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "The use of specifier here is not permit");
  }

  private static void error20(Token token)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "The type of elements in a list must the same");
  }

  private static void error21(Token token, String prodName, String altName)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "expecting " + prodName + " when " + altName + " was found");
  }

  private static void error22()
  {
    throw new RuntimeException(
      "The first production transformation must refered to the first production in section AST"
    );
  }

  private static void errorFirst(Token token)
  {
    throw new RuntimeException(
      "[" + token.getLine() + "," + token.getPos() + "] " +
      "The first production's transformation must be only one elements without an operator.");
  }

  public String toString()
  {
    StringBuffer s = new StringBuffer();
    String nl = System.getProperty("line.separator");

    s.append("ast_elems");
    s.append(nl);
    s.append(nl);

    return s.toString();
  }
}
